/*
 * org.ibex.der.* - By Brian Alliet
 * Copyright (C) 2004 Brian Alliet
 * 
 * Based on Bouncy Castle by The Legion Of The Bouncy Castle
 * Copyright (c) 2000 The Legion Of The Bouncy Castle 
 * (http://www.bouncycastle.org)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDER.S BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */
package org.ibex.crypto;
import java.io.IOException;
import java.io.*;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;
import java.math.BigInteger;

public class DER {
    public static class Null {
        final static Null instance = new Null();
        private Null() { /* noop */ }
        public boolean equals(Object o) { return o == this; }
    }

    public static class TaggedObject {
        public final Object object;
        public final int tag;
        public TaggedObject(int tag, Object object) { this.tag = tag; this.object = object; }
    }

    public static class UnknownObject {
        public final byte[] data;
        public final int tag;
        public UnknownObject(int tag,byte[] data) { this.tag = tag; this.data = data; }
    }

    public static class BitString {
        public final int paddingBits;
        public final byte[] data;
    
        public BitString(int paddingBits, byte[] data) {
            this.paddingBits = paddingBits;
            this.data = data;
        }
    }

    public static class Exception extends java.io.IOException {
        public Exception(String s) { super(s); }
    }

    public static class InputStream extends FilterInputStream {
        private static final int MAX_OBJECT_SIZE = 4*1024*1024;
    
        private int limit;
        public int bytesLeft() { return limit < 0 ? Integer.MAX_VALUE : limit-pos; }
        private int pos;
        public int getPos() { return pos; }
    
        public InputStream(java.io.InputStream is) { this(is,-1); } 
        public InputStream(java.io.InputStream is, int limit) {
            super(is);
            this.limit = limit;
        }
    
        public int read() throws IOException {
            if(limit >= 0 && pos >= limit) return -1;
            int n = super.read();
            if(n != -1) {
                pos++;
            }
            return n;
        }
    
        public int read(byte[] buf, int start, int len) throws IOException {
            if(limit >= 0) {
                if(pos >= limit) return -1;
                len = Math.min(len,limit-pos);
            }
            int n = super.read(buf,start,len);
            if(n != -1) {
                pos += n;
            }
            return n;
        }
    
        protected void readFully(byte[] buf) throws IOException {
            int pos = 0;
            int left = buf.length;
            while(left > 0) {
                int n = read(buf,pos,left);
                if(n == -1) throw new EOFException();
                pos += n;
                left -=n;
            }
        }
    
        protected int readByte() throws IOException {
            int n = read();
            if(n == -1) throw new EOFException();
            return n;
        }
    
        // From bouncycastle
        private int readLength() throws IOException {
            int length = read();
            if (length < 0) throw new IOException("EOF found when length expected");
            if (length == 0x80) return -1;      // indefinite-length encoding
            if (length > 127) {
                int size = length & 0x7f;
                length = 0;
                for (int i = 0; i < size; i++) {
                    int next = read();
                    if (next < 0) throw new IOException("EOF found reading length");    
                    length = (length << 8) + next;
                }
            }
            return length;
        }
        
        public InputStream getSequenceStream() throws IOException {
            int tag = readByte();
            int length = readLength();
            if(length < 0) throw new Exception("Indefinite length objects not supported");
            if(tag != (CONSTRUCTED|0x10)) throw new Exception("Constructed Sequence expected");
            return new InputStream(this,length);
        }
    
        private static final int CONSTRUCTED = 0x20;    
        public Object readObject() throws IOException {
            int tag = readByte();        
            int length = readLength();
            if(length < 0) throw new Exception("Indefinite length objects not supported");
            if(length > MAX_OBJECT_SIZE) throw new Exception("Object too large");
        
            switch(tag) {
                case 0x01: return buildBoolean(length); // Boolean
                case 0x02: return buildInteger(length); // Integer
                case 0x03: return buildBitString(length); // Bit String
                case 0x04: return buildOctetString(length); // Octet String
                case 0x05: return Null.instance; // NULL
                case 0x06: return buildObjectIdentifier(length); // Object Identifier
                
                case 0x13: // PrintableString
                    // It is incorrect to treat this as an IA5String but the T.61 standard is way too old and backwards
                    // to be worth supporting
                case 0x14: // T.61 String 
                case 0x16: // IA5String
                    return buildIA5String(length);
                case 0x17: return buildTime(length,false);// UTC Time
                case 0x18: return buildTime(length,true); // Generalizd Time
                
                case CONSTRUCTED | 0x10: // Constructed Sequence
                case CONSTRUCTED | 0x11: // Constructed Set
                    { 
                        return buildSequence(length);
                    }
                default: {
                    if((tag & 0x80) != 0) {
                        if ((tag & 0x1f) == 0x1f) throw new Exception("Unsupported high tag ecountered");
                        // tagged object - bottom 5 bits are tag
                        if(length == 0)
                            return new TaggedObject(tag&0x1,((tag & CONSTRUCTED) == 0) ? (Object)Null.instance : (Object)new Vector());
                        if((tag & CONSTRUCTED) == 0)
                            return new TaggedObject(tag&0x1f,buildOctetString(length));
                 
                        InputStream dis = new InputStream(this,length);
                        Object o = dis.readObject();
                        if(dis.bytesLeft() == 0) return new TaggedObject(tag&0x1f,o);
                        Vector v = new Vector();
                        v.add(o);
                        return buildSequence(dis,v);
                    } else {
                        return new UnknownObject(tag,readBytes(length));
                    }
                }     
            }
        }

        protected Vector buildSequence(int length) throws IOException { return buildSequence(new InputStream(this,length),new Vector()); }
        protected Vector buildSequence(InputStream dis,Vector v) throws IOException {
            try {
                for(;;) v.add(dis.readObject());
            } catch(EOFException e) { 
                return v; 
            }
        }    
    
        protected byte[] readBytes(int length) throws IOException {
            byte[] buf = new byte[length];
            readFully(buf);
            return buf;
        }
    
        protected BigInteger buildInteger(int length) throws IOException { return new BigInteger(readBytes(length)); }
    
        // From bouncycastle
        protected String buildObjectIdentifier(int length) throws IOException {
            byte[] bytes = readBytes(length);
            StringBuffer    objId = new StringBuffer();
            int             value = 0;
            boolean         first = true;

            for (int i = 0; i != bytes.length; i++)
                {
                    int b = bytes[i] & 0xff;

                    value = value * 128 + (b & 0x7f);
                    if ((b & 0x80) == 0)             // end of number reached
                        {
                            if (first)
                                {
                                    switch (value / 40)
                                        {
                                            case 0:
                                                objId.append('0');
                                                break;
                                            case 1:
                                                objId.append('1');
                                                value -= 40;
                                                break;
                                            default:
                                                objId.append('2');
                                                value -= 80;
                                        }
                                    first = false;
                                }

                            objId.append('.');
                            objId.append(Integer.toString(value));
                            value = 0;
                        }
                }
            return objId.toString();
        }
    
        protected String buildIA5String(int length) throws IOException {
            byte[] buf = readBytes(length);
            char[] buf2 = new char[buf.length];
            for(int i=0;i<buf.length;i++) buf2[i] = (char)(buf[i]&0xff);
            return new String(buf2);
        }
    
        private static final SimpleDateFormat dateF = new SimpleDateFormat("yyyyMMddHHmmssz");
        protected Date buildTime(int length, boolean generalized) throws IOException {
            String s = buildIA5String(length);
            if(!generalized && s.length() > 0) {
                if(s.charAt(0) < '5') s = "20" + s;
                else s = "19" + s;
            }
            switch(s.length()) {
                case 13: s = s.substring(0,12) + "00GMT+00:00"; break; //  YYYYMMDDhhmmZ
                case 15: s = s.substring(0,14) + "GMT+00:00";   break; // YYYYMMDDhhmmssZ
                case 17: s = s.substring(0,12) + "00GMT" + s.substring(12,15) + ":" + s.substring(15,17); break;  // YYYYMMDDhhmm+hh'mm'
                case 19: s = s.substring(0,14) + "GMT" + s.substring(14, 17) + ":" + s.substring(17, 19); // YYYYMMDDhhmmss+hh'mm'
                default: throw new Exception("Unknown time format " + s);
            }
            try {
                return dateF.parse(s);
            } catch(ParseException e) {
                throw new Exception("Coudln't parse time: " + e.getMessage());
            }
        }
    
        protected BitString buildBitString(int length) throws IOException {
            if(length < 1) throw new Exception("bit string too short");
            int padding = read();
            if(padding == -1) throw new IOException("unexpected eof");
            return new BitString(padding,readBytes(length-1));
        }
    
        protected byte[] buildOctetString(int length) throws IOException { return readBytes(length); }
    
        protected Boolean buildBoolean(int length) throws IOException {
            byte[] bytes = readBytes(length);
            return bytes[0] != 0 ? Boolean.TRUE : Boolean.FALSE;
        }
    
        /*
          public static void main(String[] args) throws Exception {
          InputStream is = new InputStream(new FileInputStream(args[0]));
          try {
          for(;;) dump(is.readObject(),"");
          } catch(EOFException e) {
          System.err.println("EOF");
          }
          }
          public static void dump(Object o, String indent) {
          if(o instanceof Vector) {
          Vector v = (Vector) o;
          System.out.println(indent + "Sequence/Set");
          for(int i=0;i<v.size();i++) {
          dump(v.elementAt(i),indent + i + ":  ");
          }
          } else if(o instanceof TaggedObject){
          dump(((TaggedObject)o).object,indent + "Tagged object: ");
          } else if(o instanceof byte[]) {
          System.err.println(indent + "<Byte Array>");
          } else {
          System.err.println(indent + o.toString());
          }
          }*/
    }
}
